"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.buildRouteDefinitions = exports.buildRouteDefinition = void 0;
const buildChildDefs = (route, parent, name, path) => {
    if (route.routes) {
        const toReturn = Object.assign({}, buildRouteDefinitions(route.routes, { name, path }), parent);
        return toReturn;
    }
    else {
        return parent;
    }
};
const buildDef = (route, prefix) => {
    if (route.name) {
        const name = prefix ? `${prefix.name}-${route.name}` : route.name;
        const path = prefix ? `${prefix.path}${route.path}` : route.path;
        const mine = {
            [name]: buildRouteDefinition(name, path),
        };
        return buildChildDefs(route, mine, name, path);
    }
};
/**
 * Builds a route-definition with the name and path specified.
 * Will build an array of strings with the parameters specified in the path, and
 * a reverse function -- which will substitute params in the path for values
 * passed in via an object -- to the route-definition.
 * @param path {string} - The path for this route-definition.
 * @param name {string} - The name for this route-definition.
 * @return {ZemiRouteDefinition} - A route-definition with the name, path, parameters in the path, and reverse function for said path.
 * @type{(path: string, name:string)=> ZemiRouteDefinition}
 */
function buildRouteDefinition(name, path) {
    const pathArray = path.split("/");
    const parameters = pathArray
        .filter((p) => p.includes(":"))
        .map((p) => p.split(":")[1]);
    // This is the actual reverse function that can be called within handlers
    const reverse = (paramValues) => {
        const pathParts = pathArray.map((p) => {
            if (p.includes(":")) {
                const key = p.split(":")[1];
                return paramValues[key];
            }
            else
                return p;
        });
        return pathParts.join("/");
    };
    return { name, path, parameters, reverse };
}
exports.buildRouteDefinition = buildRouteDefinition;
/**
 * Recursive. Builds a route-definition for each ZemiRoute in the array.
 * Converts zemi paths (path that has params with their types built in)
 * into valid Express paths. Used to add route-definitions to ZemiRequest
 * object.
 * @param routes {Array<ZemiRoute>} - An array of ZemiRoutes.
 * @param [prefix] {NamedRoute} - The prefix to use for this route, if nested, so that its parent's path is added to its path.
 * @returns {Record<string, ZemiRouteDefinition>} - An object where each key is a route name and the value is that routes' route-definition.
 * @type{(routes: Array<ZemiRoute>, prefix?: NamedRoute)=> Record<string, ZemiRouteDefinition>}
 */
function buildRouteDefinitions(routes, prefix) {
    const namedRoutes = routes.map((route) => buildDef(route, prefix));
    return Object.assign({}, ...namedRoutes);
}
exports.buildRouteDefinitions = buildRouteDefinitions;
//# sourceMappingURL=helpers.js.map